/*
   Class Name:

      CTerrain.

   Created by:

      Allen Sherrod (Programming Ace of www.UltimateGameProgramming.com).

   Description:

      This class is a base class used to load and save a terrain.
*/


#include"CTerrain.h"
#include<stdio.h>


CTerrain::CTerrain()
{
   // Initialize member variables.
   m_heightScale = 0.0f;
   m_vertex = NULL;
   m_color = NULL;
   m_texCoords = NULL;
   m_texCoords2 = NULL;
   m_totalVerts = 0;
   m_totalTriangles = 0;
   m_numRepeats = 1;
}


CTerrain::~CTerrain()
{
   // Release all resources.
   ShutDown();
}


bool CTerrain::LoadMap(char *filename, int detailRepeats)
{
   FILE *fp;

   // Error checking.
   if(!filename) return false;

   // If there is any map data release it.
   if(m_heightMap.mapData != NULL) ShutDown();

   // Open the file.
   fp = fopen(filename, "rb");
   if(!fp) return false;

   // Read in the size and the scale values.
   fread(&m_heightMap.size, 1, sizeof(m_heightMap.size), fp);
   fread(&m_heightScale, 1, sizeof(m_heightScale), fp);

   // Initialize the map data variable.
   m_heightMap.mapData = new unsigned char[m_heightMap.size * m_heightMap.size];

   // Error checking.
   if(m_heightMap.mapData == NULL) return false;

   // Read in the entire file into the array.
   fread(m_heightMap.mapData, 1, m_heightMap.size * m_heightMap.size, fp);

   // Number of times the detail texture map should repeat across surface.
   m_numRepeats = detailRepeats;

   CreateTerrainMesh();

   // Close the file.
   fclose(fp);
   return true;
}


void CTerrain::CreateTerrainMesh()
{
   // Index counters.
   int i = 0, c = 0, t = 0, t2 = 0;

   if(m_color) delete[] m_color;
   if(m_vertex) delete[] m_vertex;
   if(m_texCoords) delete[] m_texCoords;
   if(m_texCoords2) delete[] m_texCoords2;

   m_totalVerts = 0;
   m_totalTriangles = 0;

   //Sizes we need for the points/color and tex coords.
   int size = ((m_heightMap.size - 1) *
              (m_heightMap.size - 1) * 6) * 3;
   int tSize = ((m_heightMap.size - 1) *
              (m_heightMap.size - 1) * 6) * 2;

   // Create enough space for all vertex points and their colors.
   m_vertex = new float[size];
   m_color = new float[size];
   m_texCoords = new float[tSize];
   m_texCoords2 = new float[tSize];

   // Loop through and generate a grid.  Use the height map
   // when setting the Y to create the terrain mesh.  Create 2 triangles
   // each iteration.
   for(int z = 0; z < m_heightMap.size - 1; z++)
	   {
			for(int x = 0; x < m_heightMap.size - 1; x++)
			   {
			      // Calculate texture coords for these tris.
			      float left = (float)x / m_heightMap.size;
			      float right = ((float)x + 1) / m_heightMap.size;
				   float bottom = (float)z / m_heightMap.size;
				   float top = (float)(z + 1) / m_heightMap.size;

               // V1.
			      m_texCoords[t++] = left;
			      m_texCoords[t++] = bottom;
			      m_texCoords2[t2++] = left * m_numRepeats;
			      m_texCoords2[t2++] = bottom * m_numRepeats;
			      m_color[c++] = GetHeight(x, z);
			      m_color[c++] = GetHeight(x, z);
			      m_color[c++] = GetHeight(x, z);
			      m_vertex[i++] = (float)x;
			      m_vertex[i++] = GetScaledHeight(x, z);
			      m_vertex[i++] = (float)z;
			      m_totalVerts++;

               // V2.
               m_texCoords[t++] = left;
			      m_texCoords[t++] = top;
			      m_texCoords2[t2++] = left * m_numRepeats;
			      m_texCoords2[t2++] = top * m_numRepeats;
               m_color[c++] = GetHeight(x, z + 1);
               m_color[c++] = GetHeight(x, z + 1);
               m_color[c++] = GetHeight(x, z + 1);
			      m_vertex[i++] = (float)x;
			      m_vertex[i++] = GetScaledHeight(x, z + 1);
			      m_vertex[i++] = (float)z + 1;
			      m_totalVerts++;

               // V3.
               m_texCoords[t++] = right;
			      m_texCoords[t++] = bottom;
			      m_texCoords2[t2++] = right * m_numRepeats;
			      m_texCoords2[t2++] = bottom * m_numRepeats;
               m_color[c++] = GetHeight(x + 1, z);
               m_color[c++] = GetHeight(x + 1, z);
               m_color[c++] = GetHeight(x + 1, z);
			      m_vertex[i++] = (float)x + 1;
			      m_vertex[i++] = GetScaledHeight(x + 1, z);
			      m_vertex[i++] = (float)z;
			      m_totalVerts++;


               // V4.
               m_texCoords[t++] = right;
			      m_texCoords[t++] = bottom;
			      m_texCoords2[t2++] = right * m_numRepeats;
			      m_texCoords2[t2++] = bottom * m_numRepeats;
               m_color[c++] = GetHeight(x + 1, z);
               m_color[c++] = GetHeight(x + 1, z);
               m_color[c++] = GetHeight(x + 1, z);
			      m_vertex[i++] = (float)x + 1;
			      m_vertex[i++] = GetScaledHeight(x + 1, z);
			      m_vertex[i++] = (float)z;
			      m_totalVerts++;

               // V5.
               m_texCoords[t++] = right;
			      m_texCoords[t++] = top;
			      m_texCoords2[t2++] = right * m_numRepeats;
			      m_texCoords2[t2++] = top * m_numRepeats;
               m_color[c++] = GetHeight(x + 1, z + 1);
               m_color[c++] = GetHeight(x + 1, z + 1);
               m_color[c++] = GetHeight(x + 1, z + 1);
			      m_vertex[i++] = (float)x + 1;
			      m_vertex[i++] = GetScaledHeight(x + 1, z + 1);
			      m_vertex[i++] = (float)z + 1;
			      m_totalVerts++;

               // V6.
               m_texCoords[t++] = left;
			      m_texCoords[t++] = top;
               m_texCoords2[t2++] = left * m_numRepeats;
			      m_texCoords2[t2++] = top * m_numRepeats;
               m_color[c++] = GetHeight(x, z + 1);
               m_color[c++] = GetHeight(x, z + 1);
               m_color[c++] = GetHeight(x, z + 1);
			      m_vertex[i++] = (float)x;
			      m_vertex[i++] = GetScaledHeight(x, z + 1);
			      m_vertex[i++] = (float)z + 1;
			      m_totalVerts++;

               m_totalTriangles += 2;
			   }
	   }

   // Convert between 0 and 1.
	for(int k = 0; k < size; k++)
      {
         if(m_color[k] != 0) m_color[k] /= 255;
      }
}


bool CTerrain::SaveMap(char *filename)
{
	FILE *fp;

   // Error Checking...
   if(m_heightMap.mapData == 0) return false;
   if(filename == 0) return false;

	// Open the file to write it out as an text file.
	fp = fopen(filename, "wb");
	
   // Error checking.
   if(fp == 0) return false;

   // Write out the size and scale values.
   fwrite(&m_heightMap.size, 1, sizeof(m_heightMap.size), fp);
   fwrite(&m_heightScale, 1, sizeof(m_heightScale), fp);

	// Write out all of the data in one big write.
	fwrite(m_heightMap.mapData, 1, m_heightMap.size * m_heightMap.size, fp);
	
	// Close the file.
	fclose(fp);
	return true;
}


void CTerrain::SetHeightScale(float val)
{
   // Set the scale value to want is passed in.
   m_heightScale = val;
}


void CTerrain::SetHeight(unsigned char val, int x, int z)
{
   // Set the value to the specified location.
   m_heightMap.mapData[x + m_heightMap.size * z] = val;
}


unsigned char CTerrain::GetHeight(int x, int z)
{
   // Return the unscaled height value of the specified point.
   return m_heightMap.mapData[x + m_heightMap.size * z];
}


float CTerrain::GetScaledHeight(int x, int z)
{
   // Return the scaled height value of the specified point.
   return (m_heightMap.mapData[x + m_heightMap.size * z] * m_heightScale);
}


float CTerrain::RegionPercent(int tileType, unsigned char height)
{
	float distance1 = 0, distance2 = 0;

   // First we test each tile and see if the test height is greater than its optimal height.
   // We only test those that are the type we sent it.
	if(m_tiles.textureTiles[LOWEST_TILE].image)
	   {
		   if(tileType == LOWEST_TILE && height < m_tiles.regions[LOWEST_TILE].optimalHeight)
			   return 1.0f;
	   }
	else if(m_tiles.textureTiles[LOW_TILE].image)
	   {
		   if(tileType == LOW_TILE && height < m_tiles.regions[LOW_TILE].optimalHeight)
			   return 1.0f;
	   }
	else if(m_tiles.textureTiles[HIGH_TILE].image)
	   {
		   if(tileType == HIGH_TILE && height < m_tiles.regions[HIGH_TILE].optimalHeight)
			   return 1.0f;
	   }
	else if(m_tiles.textureTiles[HIGHEST_TILE].image)
	   {
		   if(tileType == HIGHEST_TILE && height < m_tiles.regions[HIGHEST_TILE].optimalHeight)
			   return 1.0f;
	   }

	// If the height we are testing is below or above the choose tile type then we return 0.
	if(height < m_tiles.regions[tileType].lowHeight)
		return 0.0f;
	else if(height > m_tiles.regions[tileType].highHeight)
		return 0.0f;

	// If the height is below the optimal height
	if(height < m_tiles.regions[tileType].optimalHeight)
	   {
		   // Get the distance between the m_tiles lowest height and the test height.
		   distance1 = (float)(height - m_tiles.regions[tileType].lowHeight);

         // Get the distance between the optimal height and the lowest height.
		   distance2 = (float)(m_tiles.regions[tileType].optimalHeight -
		               m_tiles.regions[tileType].lowHeight);

         // return the percent.
		   return (distance1 / distance2);
	   }
   else if(height == m_tiles.regions[tileType].optimalHeight)
      {
         // Else if the height equals the optimal height we send in full color.
		   return 1.0f;
      }
	else if(height > m_tiles.regions[tileType].optimalHeight)
	   {
		   // Else if the height is greater than the optimal height.
         // Get the distance between the max height from the optimal height.
		   distance1 = (float)(m_tiles.regions[tileType].highHeight -
                         m_tiles.regions[tileType].optimalHeight);

         // Return the distance - the test height -
         // optimal height / by the distance to get %.
		   return((distance1 - (height -
		          m_tiles.regions[tileType].optimalHeight)) / distance1);
	   }

   // Return 0 if we get here.
	return 0.0f;
}


void CTerrain::GetTexCoords(unsigned int texWidth, unsigned int texHeight,
                            unsigned int x, unsigned int z, float &tu, float &tv)
{
   // These two will store the amount of times the texture
   // repeats across the terrain along the x and z.
	int totalWidthRepeats = -1;
	int totalHeightRepeats = -1;

   // Temp number to hold temp values.
   int i = 0;

	// Calculate the amount of times this image would repeat across the x.
	while(totalWidthRepeats == -1)
	   {
		   i++;

		   // We test if this point (x value) is less than the width times i.
         // If x is 156 and the image width is 256 then it will not repeat.
         // If x is 512 and the image width is 256 then the image will repeat
         // at least 1 time.
		   if(x < (texWidth * i)) totalWidthRepeats = i - 1;
	   }

	// Reset.
	i = 0;

	// Calculate the amount of times this image would repeat across the z.
	while(totalHeightRepeats == -1)
	   {
		   i++;

		   // Same as with the x.
		   if(z < (texHeight * i)) totalHeightRepeats = i - 1;
	   }

	// Calculate the final texture coordinate.
	tu = (float)(x - (texWidth * totalWidthRepeats));
	tv = (float)(z - (texHeight * totalHeightRepeats));
}


unsigned char CTerrain::InterpolateHeight(int x, int z, float textureMapRatio)
{
	unsigned char low = 0, high = 0;
	float interpolatedX = 0.0f, interpolatedZ = 0.0f;
	float interpolation = 0.0f;
   float scaledX = x * textureMapRatio;
	float scaledZ = z * textureMapRatio;

	// Set low as the height of this point.
	low = GetHeight((int)scaledX, (int)scaledZ);

	// Set the high value.  If scaled x + 1 is greater than the entire size of the map then
   // we send the low to return as default.  Else we grab the high x.
   // This way we don't go out  of bounds in the array.
	if((scaledX + 1) > m_heightMap.size)
		return low;
	else
		high = GetHeight((int)scaledX + 1, (int)scaledZ);

	// Set the interpolation to be the remainder of scaled x - itself.
	// This will give us a value between  0 and 0.99.  Then we interpolate
	// the high and low.
	interpolation = (scaledX - (int)scaledX);
	interpolatedX = ((high - low) * interpolation) + low;

	// Next we do the same thing for the z that we did for the x.
	if((scaledZ + 1 ) > m_heightMap.size)
		return low;
	else
		high = GetHeight((int)scaledX, (int)scaledZ + 1);

	// Calculate the interpolation for z.
	interpolation = (scaledZ - (int)scaledZ);
	interpolatedZ = ((high - low) * interpolation) + low;

	// Average out the interpolation for the x and z.
	return ((unsigned char)((interpolatedX + interpolatedZ) / 2));
}


unsigned char *CTerrain::GenerateTextureMap(unsigned int imageSize)
{
   // Red, green, and blue color values.  OpenGL texture id.
	unsigned char red = 0, green = 0, blue = 0;

   // Set of texture coords and last height which is used in combining the images.
	float tu, tv;
   int lastHeight = -1;

   // Each of these hold the total color values of each used in the final image.
	float totalRed = 0, totalGreen = 0, totalBlue = 0;
	
   // Blend amounts for each texture tile and the ratio between the terrain size and image size.
   float blendList[MAX_TILES] = {0};
	float textureMapRatio = (float)m_heightMap.size / (float)imageSize;

	// Reset the number of m_tiles.
	m_tiles.numTiles = 0;

   // First we loop through the max amount of m_tiles we can have and see how many actually
   // have images.  This way the user does not have to have a set number of images before
   // we can generate a texture.  Need just at least 1 to see anything.
	for(int i = 0; i < MAX_TILES; i++)
	   {
		   if(m_tiles.textureTiles[i].image) m_tiles.numTiles++;
	   }

   if(m_tiles.numTiles == 0) return NULL;

	// Loop through again to calculate the tile regions.
	for(i = 0; i < MAX_TILES; i++)
	   {
		   // If the image was loaded, perform calculations.
		   if(m_tiles.textureTiles[i].image)
		      {
			      // Here we calculate the low, optimal, and high height for this region.
			      m_tiles.regions[i].lowHeight = lastHeight + 1;
   			   
               lastHeight += 255 / m_tiles.numTiles;

			      m_tiles.regions[i].optimalHeight = lastHeight;

			      m_tiles.regions[i].highHeight = (lastHeight -
			         m_tiles.regions[i].lowHeight) + lastHeight;
		      }
	   }

	// Create the texture array.
	unsigned char *image = new unsigned char[imageSize * imageSize * 3];
	
	// And the last part in the generation is to create the 2D image.
	for(unsigned int z = 0; z < imageSize; z++)
	   {
		   for(unsigned int x = 0; x < imageSize; x++)
		      {
			      // Intialize the color variables.
			      totalRed = 0.0f;
			      totalGreen = 0.0f;
			      totalBlue = 0.0f;

			      // Loop through all tiles and generate each final pixel color.
			      for(i = 0; i < m_tiles.numTiles; i++)
			         {
					      // Calculate the texture coords.  This is used to
					      // get a pixel from  each tile.
					      GetTexCoords(m_tiles.textureTiles[i].imageWidth,
                                 m_tiles.textureTiles[i].imageHeight, x, z,
                                 tu, tv);

					      // Get the color of the pixel for the set of texture coords.
					      int index = (((unsigned int)tv * imageSize) + (unsigned int)tu) * 3;
					      red = m_tiles.textureTiles[i].image[index + 0];
					      green = m_tiles.textureTiles[i].image[index + 1];
					      blue = m_tiles.textureTiles[i].image[index + 2];
         
					      // Calculate the amount to blend this pixel with the rest.
					      blendList[i] = RegionPercent(i, InterpolateHeight(x, z,
					                                       textureMapRatio));

					      // Store the pixel from this tile to the total.
					      totalRed += red * blendList[i];
					      totalGreen += green * blendList[i];
					      totalBlue += blue * blendList[i];
			         }

			      // Set a pixel in our final texture image.
			      image[((z * imageSize) + x) * 3] = (unsigned char)totalRed;
               image[((z * imageSize) + x) * 3 + 1] = (unsigned char)totalGreen;
               image[((z * imageSize) + x) * 3 + 2] = (unsigned char)totalBlue;
		      }
	   }

   return image;
}


void CTerrain::ShutDown()
{
   // Release all resources used by the terrain.
   m_heightScale = 0.0f;
   m_tiles.textureTiles[0].FreeImage();
   m_tiles.textureTiles[1].FreeImage();
   m_tiles.textureTiles[2].FreeImage();
   m_tiles.textureTiles[3].FreeImage();
   
   if(m_vertex)
      {
         delete[] m_vertex;
         m_vertex = NULL;
      }
   
   if(m_color)
      {
         delete[] m_color;
         m_color = NULL;
      }

   if(m_texCoords)
      {
         delete[] m_texCoords;
         m_texCoords = NULL;
      }

   if(m_texCoords2)
      {
         delete[] m_texCoords2;
         m_texCoords2 = NULL;
      }

   if(m_heightMap.mapData)
      {
         delete[] m_heightMap.mapData;
         m_heightMap.mapData = NULL;
         m_heightMap.size = 0;
      }
}


// Copyright July 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com