/*
   Class Name:

      CTerrain.

   Created by:

      Allen Sherrod (Programming Ace of www.UltimateGameProgramming.com).

   Description:

      This class is a base class used to load and save a terrain.
*/


#ifndef _CTERRAIN_H_
#define _CTERRAIN_H_

#include"CTgaLoader.h"

#define MAX_TILES 5


class CHeightMap
{
   public:
      CHeightMap() { mapData = 0; size = 0; }
      ~CHeightMap()
         {
            size = 0;
            if(mapData) delete[] mapData;
            mapData = 0;
         }

      unsigned char *mapData;
      int size;
};


enum TILE_TYPES
{
	LOWEST_TILE = 0,        // Dirt and anything under grass.
	LOW_TILE,               // Grass and stuff of that nature.
	HIGH_TILE,              // Moutain sides.
	HIGHEST_TILE            // Top of mountains (snow, etc).
};


struct TextureRegions
{
	int lowHeight;          // Lowest height.
	int optimalHeight;      // Ideal height.
	int highHeight;         // Highest possible height.
};


struct TerrainTiles
{
	TextureRegions regions[MAX_TILES];  // Texture regions.
	CTGALoader textureTiles[MAX_TILES];	// Texture tiles.
	int numTiles;                       // Total number of tiles.
};


class CTerrain
{
   public:
      CTerrain();
      ~CTerrain();

      bool LoadMap(char *filename, int detailRepeats);
      bool SaveMap(char *filename);
      void CreateTerrainMesh();

      void SetHeightScale(float val);
      void SetHeight(unsigned char val, int x, int z);
      unsigned char GetHeight(int x, int z);
      float GetScaledHeight(int x, int z);

      float RegionPercent(int tileType, unsigned char height);
      void GetTexCoords(unsigned int texWidth, unsigned int texHeight, unsigned int x,
                        unsigned int z, float &tu, float &tv);
      unsigned char InterpolateHeight(int x, int z, float textureMapRatio);
      unsigned char *GenerateTextureMap(unsigned int imageSize);

      inline bool LoadTile(TILE_TYPES tileType, char *filename)
	      {
            return m_tiles.textureTiles[tileType].LoadTGA(filename);
         }

      void ShutDown();

      // Stores height map and scale value.
      CHeightMap m_heightMap;
      float m_heightScale;

      // A list of images used to generate terrain texture.
      TerrainTiles m_tiles;

      // Vertex list and vertex colors.
      float *m_vertex;
      float *m_color;
      float *m_texCoords;
      float *m_texCoords2;

      int m_totalVerts;
      int m_totalTriangles;
      int m_numRepeats;
};

#endif


// Copyright July 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com