/*
   Texture Generation Terrain Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to generate the texture that will be
   mapped to the terrain by combining a bunch of smaller images.
   This is useful especially if you are generating the terrain.
   Even if the terrain is not being generated dynamically this
   can help you create a surface for thet terrain that looks
   great without the need for expensive tools that are used
   to texture meshes.  Plus if you change the terrain you don't
   have to alter the texture to match.
*/


// Demo: Texture Generation Terrain.


#define WIN32_LEAN_AND_MEAN// Trims down the libraries.
#define VC_LEANMEAN        // Trims even further.


#include<windows.h>
#include<gl/gl.h>
#include<gl/glu.h>
#include<gl/glext.h>
#include"CTerrain.h"
#include"CTgaLoader.h"


// Function prototypes...
bool InitializeGL();
void RenderScene();
void Shutdown();


// Device context.
HDC g_HDC;

// xRotation and yRotation is used to rotate the entire scene.
float xRotation = 0.0f;
float yRotation = 0.0f;

// Terrain object.
CTerrain Terrain;
unsigned int terrainTexID = -1;
CTGALoader terrainDetailTex;

// Multitexture function.
PFNGLACTIVETEXTUREARBPROC glActiveTextureARB = NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC glClientActiveTextureARB = NULL;


void SetupPixelFormat(HDC hDC)
{
   int nPixelFormat;

   static PIXELFORMATDESCRIPTOR pfd = {
         sizeof(PIXELFORMATDESCRIPTOR),   // size of structure.
         1,                               // always 1.
         PFD_DRAW_TO_WINDOW |             // support window
         PFD_SUPPORT_OPENGL |             // support OpenGl
         PFD_DOUBLEBUFFER,                // support double buffering
         PFD_TYPE_RGBA,                   // support RGBA
         16,                              // bit color mode
         0, 0, 0, 0, 0, 0,                // ignore color bits
         0,                               // no alpha buffer
         0,                               // ignore shift bit
         0,                               // no accumulation buffer
         0, 0, 0, 0,                      // ignore accumulation bits.
         16,                              // number of depth buffer bits.
         0,                               // number of stencil buffer bits.
         0,                               // 0 means no auxiliary buffer
         PFD_MAIN_PLANE,                  // The main drawing plane
         0,                               // this is reserved
         0, 0, 0 };                       // layer masks ignored.

   // This chooses the best pixel format and returns index.
   nPixelFormat = ChoosePixelFormat(hDC, &pfd);

   // This set pixel format to device context.
   SetPixelFormat(hDC, nPixelFormat, &pfd);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static HGLRC hRC;                         // Rendering context.
   static HDC hDC;                           // Device context.
   int width, height;                        // The window width and height.
	static POINT oldMousePos;                 // Last mouse position.
   static POINT currentMousePos;             // Current mouse position.
   static bool isMouseActive;                // Is the left mouse button down.

   switch(message)
      {
         case WM_CREATE:                     // Windows creation.
            hDC = GetDC(hwnd);               // This gets the device context for our window.
            g_HDC = hDC;                     // Assigns the global device context to this one.
            SetupPixelFormat(hDC);           // Call the pixel format function.

            hRC = wglCreateContext(hDC);     // Creates the rendering context.
            wglMakeCurrent(hDC, hRC);        // Makes the rendering context.
            return 0;
            break;


         case WM_CLOSE:                      // Close message.
         case WM_DESTROY:
            wglMakeCurrent(hDC, NULL);
            wglDeleteContext(hRC);           // Deletes the rendering context.

            PostQuitMessage(0);              // Says close the program.
            return 0;
            break;


         case WM_SIZE:                       // re-size message.
            height = HIWORD(lParam);         // This gets the height of the window.
            width = LOWORD(lParam);          // This gets the width of the window.

            if(height==0)                    // we don't want it to be possible for a
               {                             // height of 0.  If it is 0 me make it 1.
                  height = 1;
               }

            glViewport(0, 0, width, height); // resets the viewport to new dimensions.
            glMatrixMode(GL_PROJECTION);     // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            // Calculate the aspect ratio of the window.
            gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 1000.0f);

            glMatrixMode(GL_MODELVIEW);      // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            return 0;
            break;


         case WM_KEYDOWN:
		      switch(wParam)
               { 
                  // Close the app if the user presses escape.
		            case VK_ESCAPE:
                     PostQuitMessage(0);
				         break;

                  default:
                     break;
               }
            break;


			case WM_LBUTTONDOWN:
				oldMousePos.x = currentMousePos.x = LOWORD (lParam);
				oldMousePos.y = currentMousePos.y = HIWORD (lParam);
            isMouseActive = true;
			   break;


			case WM_LBUTTONUP:
				isMouseActive = false;
				break;


			case WM_MOUSEMOVE:
				currentMousePos.x = LOWORD (lParam);
				currentMousePos.y = HIWORD (lParam);

				if(isMouseActive)
					{
						xRotation -= (currentMousePos.x - oldMousePos.x);
						yRotation -= (currentMousePos.y - oldMousePos.y);
					}

				oldMousePos.x = currentMousePos.x;
				oldMousePos.y = currentMousePos.y;
			   break;


         default:
            break;
      }

   // What this does is pass all of the unhandled messages to DefWindowProc.
   return (DefWindowProc(hwnd, message, wParam, lParam));
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                        // A message variable.
   WNDCLASSEX windowClass;                         // Your Window class.
   HWND hwnd;                                      // The Window handle.
   bool isFinished;                                // True then exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);        // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;    // style of the window.
   windowClass.lpfnWndProc = WndProc;              // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                     // Extra class information.
   windowClass.cbWndExtra = 0;                     // Extra window information.
   windowClass.hInstance = hInstance;              // Handle of application Instance.
   // Handle of application Icon.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);// mouse cursor
   windowClass.hbrBackground = NULL;               // background color.
   windowClass.lpszMenuName = NULL;                // name of the main menu.
   windowClass.lpszClassName = "UltimateGameProgrammingClass";// window class name.
   // icon when minimized.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create window...
   hwnd = CreateWindowEx(NULL,// The extended window style.
                        "UltimateGameProgrammingClass",// window Class name.
                        "Texture Generation Demo - By the Programming Ace.",// window name.
                        WS_OVERLAPPEDWINDOW | WS_VISIBLE |// The window style.
                        WS_SYSMENU | WS_CLIPCHILDREN |// window style.
                        WS_CLIPSIBLINGS,// window style.
                        100, 100,// window x, y coordinate.
                        640, 480,// window width and height.
                        NULL,// handle to parent window.
                        NULL,// menu.
                        hInstance,// handle to app instance.
                        NULL);   // pointer to window creation data.
   
   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW); // This shows the window.
   UpdateWindow(hwnd);        // This forces a paint message.

   isFinished = false;        // False = running, True = not running.

   // If initialize fail (return false), then we don't want the program to run.
   if(!InitializeGL()) isFinished = true;

   // This is the messsage loop.
   while(!isFinished)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message then quit the app.
               if(msg.message == WM_QUIT)
                  {
                     isFinished = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

   // Shutdown (free) all resources we used up.
   Shutdown();

   return (int)msg.wParam;
}


bool InitializeGL()
{
   glClearColor(0.6f, 0.6f, 1.0f, 1.0f);

   glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE_2D);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

   // With this we will load the list of extensions the hardware supports.
   char *extension = (char*)glGetString(GL_EXTENSIONS);
   if(strstr(extension, "GL_ARB_multitexture") == 0) return false;

   // Load the multitexture funtion and test if all went well.
   glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC)wglGetProcAddress("glActiveTextureARB");
   glClientActiveTextureARB = (PFNGLCLIENTACTIVETEXTUREARBPROC)wglGetProcAddress("glClientActiveTextureARB");
   if(!glActiveTextureARB || !glClientActiveTextureARB) return false;


   // Load map.
   unsigned char *image = NULL;
   int imgSize = 0;

   if(!Terrain.LoadMap("map.raw", 8)) return false;

   // Load our textures for each level.  They will be blended into one huge texture
   // based on the height data of the terrain.
   if(!Terrain.LoadTile(LOWEST_TILE, "lowest.tga")) return false;
	if(!Terrain.LoadTile(LOW_TILE, "low.tga")) return false;
	if(!Terrain.LoadTile(HIGH_TILE, "high.tga")) return false;
	if(!Terrain.LoadTile(HIGHEST_TILE, "highest.tga")) return false;

   // Generate the terrain texture.
   image = Terrain.GenerateTextureMap(256);
   if(!image) return false;

   // Load both textures.
   glGenTextures(1, &terrainTexID);
	glBindTexture(GL_TEXTURE_2D, terrainTexID);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);						
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 256, 256,
	             0, GL_RGB, GL_UNSIGNED_BYTE, image);

   delete[] image;

   if(!terrainDetailTex.LoadTGA("detail.tga")) return false;
   glGenTextures(1, &terrainDetailTex.ID);
	glBindTexture(GL_TEXTURE_2D, terrainDetailTex.ID);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);						
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, terrainDetailTex.imageWidth,
	             terrainDetailTex.imageHeight, 0, GL_RGB, GL_UNSIGNED_BYTE,
	             terrainDetailTex.image);

   return true;
}


void RenderScene()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   // Clears the screen.
   glLoadIdentity();                   //Reset modelview matrix for new frame.

	glTranslatef(-50.0f, -50.0f, -150);
	glRotatef(-yRotation, 1.0f, 0.0f, 0.0f);
	glRotatef(-xRotation, 0.0f, 1.0f, 0.0f);

   glActiveTextureARB(GL_TEXTURE0_ARB);
   glEnable(GL_TEXTURE_2D);
   glBindTexture(GL_TEXTURE_2D, terrainTexID);
   
   glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_2D);
   glBindTexture(GL_TEXTURE_2D, terrainDetailTex.ID);
   
   glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
	glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 2);

   // Enable all client states we are using.
   glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	
	// Load the data to each pointer type we need.
	glVertexPointer(3, GL_FLOAT, 0, Terrain.m_vertex);
	glTexCoordPointer(2, GL_FLOAT, 0, Terrain.m_texCoords);
	
	glClientActiveTextureARB(GL_TEXTURE1_ARB);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer(2, GL_FLOAT, 0, Terrain.m_texCoords2);

	glClientActiveTextureARB(GL_TEXTURE0_ARB);
	
	// Draw the entire object.
	glDrawArrays(GL_TRIANGLES, 0, Terrain.m_totalVerts);

   // Disable all the client states we enabled.
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTextureARB(GL_TEXTURE1_ARB);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTextureARB(GL_TEXTURE0_ARB);

	SwapBuffers(g_HDC);
}


void Shutdown()
{
   // Release all resources our terrain uses.
   Terrain.ShutDown();
   terrainDetailTex.FreeImage();
   glDeleteTextures(1, &terrainTexID);
}


// Recap:

// Generating a texture across a terrain can help a lot.  By
// making different base tiles you can get different results that
// are mapped to the terrain accurately.  Plus on top of that
// if you are generating the terrain, then generating the texture
// is the only thing you can do if you want it to look right.


// Copyright July 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com